/// @file   RFIO.cpp
/// @brief  RFIO IO.
/// @author CERN DPM <dpm-devel@cern.ch>
#include <dmlite/cpp/dmlite.h>
#include <dpm_api.h>
#include <errno.h>
#include <rfio_api.h>
#include <serrno.h>
#include "Adapter.h"
#include "DpmAdapter.h"
#include "FunctionWrapper.h"
#include "RFIO.h"

using namespace dmlite;

// Prototypes of Cthread methods (not in the headers)
extern "C" {
int Cthread_init(void);
int _Cthread_addcid(char *, int, char *, int, pthread_t, unsigned, void *(*)(void *), int);
int rfio_parse(char *, char **, char **);
}



StdRFIOFactory::StdRFIOFactory(): passwd_("default"), useIp_(true)
{
  Cthread_init();
  _Cthread_addcid(NULL, 0, NULL, 0, pthread_self(), 0, NULL, 0);
  
  setenv("CSEC_MECH", "ID", 1);
  dpm_client_resetAuthorizationId();
}



StdRFIOFactory::~StdRFIOFactory()
{
  // Nothing
}



void StdRFIOFactory::configure(const std::string& key, const std::string& value) throw (DmException)
{
  if (key == "TokenPassword") {
    this->passwd_ = value;
  }
  else if (key == "TokenId") {
    if (strcasecmp(value.c_str(), "ip") == 0)
      this->useIp_ = true;
    else
      this->useIp_ = false;
  }
  else if (key == "DpmHost" || key == "Host") {
    setenv("DPM_HOST", value.c_str(), 1);
    setenv("DPNS_HOST", value.c_str(), 1);
  }
  else
    throw DmException(DMLITE_CFGERR(DMLITE_UNKNOWN_KEY),
                      key + " not known");
}



IODriver* StdRFIOFactory::createIODriver(PluginManager* pm) throw (DmException)
{
  return new StdRFIODriver(this->passwd_, this->useIp_);
}



StdRFIODriver::StdRFIODriver(std::string passwd, bool useIp):
  si_(0), secCtx_(0), passwd_(passwd), useIp_(useIp)
{
  dpm_client_resetAuthorizationId();
}



StdRFIODriver::~StdRFIODriver()
{
  // Nothing
}



std::string StdRFIODriver::getImplId() const throw ()
{
  return "StdRFIODriver";
}



void StdRFIODriver::setStackInstance(StackInstance* si) throw (DmException)
{
  dpm_client_resetAuthorizationId();
  this->si_ = si;
}



void StdRFIODriver::setSecurityContext(const SecurityContext* ctx) throw (DmException)
{
  this->secCtx_ = ctx;
}



IOHandler* StdRFIODriver::createIOHandler(const std::string& pfn,
                                          int flags,
                                          const Extensible& extras,
                                          mode_t mode) throw (DmException)
{
  if ( !(flags & IODriver::kInsecure) ) {
      if (!extras.hasField("token"))
        throw DmException(EACCES, "Missing token");

      std::string userId;
      if (this->useIp_)
        userId = this->secCtx_->credentials.remoteAddress;
      else
        userId = this->secCtx_->credentials.clientName;

      if (dmlite::validateToken(extras.getString("token"),
                                userId,
                                pfn, this->passwd_,
                                flags != O_RDONLY) != kTokenOK)
        throw DmException(EACCES, "Token does not validate (using %s)",
                          this->useIp_?"IP":"DN");
  }
  
  // Create
  return new StdRFIOHandler(pfn, flags, mode);
}

StdRFIOHandler::pp::pp(int fd, bool *peof, off_t np):
  fd(fd), peof(peof)
{
  if ((pos = rfio_lseek64(fd, 0, SEEK_CUR)) == -1)
    throw DmException(serrno, "Could not seek");
  if (rfio_lseek64(fd, np, SEEK_SET) == - 1)
    throw DmException(serrno, "Could not seek");
  eof = *peof; *peof = false;
}

StdRFIOHandler::pp::~pp()
{
  if (rfio_lseek64(fd, pos, SEEK_SET) == - 1)
    throw DmException(serrno, "Could not seek");
  *peof = eof;
}

void StdRFIODriver::doneWriting(const Location& loc) throw (DmException)
{
  struct dpm_filestatus     *statuses;
  int                        nReplies;
  std::string                sfn;
  
  if (loc.empty())
    throw DmException(EINVAL, "Empty location");

  // Need the sfn
  sfn = loc[0].url.query.getString("sfn");
  if (sfn.empty())
    throw DmException(EINVAL, "sfn not specified");

  // Need the dpm token
  std::string token = loc[0].url.query.getString("dpmtoken");
  if (token.empty())
    throw DmException(EINVAL, "dpmtoken not specified"); 

  const char* sfnPtr[] = { sfn.c_str() };
  FunctionWrapper<int, char*, int, char**, int*, dpm_filestatus**>
    (dpm_putdone, (char*)token.c_str(), 1, (char**)&sfnPtr, &nReplies, &statuses)(3);

  dpm_free_filest(nReplies, statuses);
}



StdRFIOHandler::StdRFIOHandler(const std::string& path,
                               int flags, mode_t mode) throw (DmException):
  eof_(false), islocal_(false)
{ 
  int err;
  char *host, *filename;
  if ((err=pthread_mutex_init(&this->mtx_, 0)))
    throw DmException(err, "Could not create a new mutex");
  if (!rfio_parse((char *)path.c_str(),&host,&filename) && !host)
    this->islocal_ = true;
  this->fd_ = rfio_open64((char *)path.c_str(), flags, mode);
  if (this->fd_ == -1)
    throw DmException(rfio_serrno(), "Could not open %s", path.c_str());
}



StdRFIOHandler::~StdRFIOHandler()
{
  if (this->fd_ != -1)
    rfio_close(this->fd_);
  pthread_mutex_destroy(&this->mtx_);
}



void StdRFIOHandler::close() throw (DmException)
{
  rfio_close(this->fd_);
  this->fd_ = -1;
}



size_t StdRFIOHandler::read(char* buffer, size_t count) throw (DmException)
{
  lk l(islocal_ ? 0 : &this->mtx_);
  size_t nbytes = rfio_read(this->fd_, buffer, count);
  
  eof_  = (nbytes < count);
  
  return nbytes;
}

size_t StdRFIOHandler::pread(void* buffer, size_t count, off_t offset) throw (DmException)
{
  if (this->islocal_)
    return static_cast<size_t>(::pread(this->fd_, buffer, count, offset));
  lk l(&this->mtx_);
  pp p(this->fd_, &this->eof_, offset);
  return rfio_read(this->fd_, buffer, count);
}

size_t StdRFIOHandler::write(const char* buffer, size_t count) throw (DmException)
{
  lk l(islocal_ ? 0 : &this->mtx_);
  size_t nbytes = rfio_write(this->fd_, (char *)buffer, count);
  return nbytes;
}

size_t StdRFIOHandler::pwrite(const void* buffer, size_t count, off_t offset) throw (DmException)
{
  if (this->islocal_)
    return static_cast<size_t>(::pwrite(this->fd_, buffer, count, offset));
  lk l(&this->mtx_);
  pp p(this->fd_, &this->eof_, offset);
  return rfio_write(this->fd_, (char *)buffer, count);
}

void StdRFIOHandler::seek(off_t offset, Whence whence) throw (DmException)
{
  lk l(islocal_ ? 0 : &this->mtx_);
  if (rfio_lseek64(this->fd_, offset, whence) == -1)
    throw DmException(serrno, "Could not seek");
}



off_t StdRFIOHandler::tell(void) throw (DmException)
{
  lk l(islocal_ ? 0 : &this->mtx_);
  return rfio_lseek64(this->fd_, 0, SEEK_CUR);
}



void StdRFIOHandler::flush(void) throw (DmException)
{
  // Nothing
}



bool StdRFIOHandler::eof(void) throw (DmException)
{
  return eof_;
}
